/*
  ######################################################################
  # THIS IS THE STRIPPED VERSION OF THE ORIGINAL FILE                  #
  ######################################################################
  # This file was taken from Lua 5.1.1 tar-ball distribution           #
  ######################################################################
  # June 20, 2006 - Aloysius Indrayanto                                #
  # All comments (except the library copyright & license) are stripped #
  ######################################################################
  # See the VortexGE documentation for information about VortexGE      #
  ######################################################################
*/

/*
** $Id: llex.h,v 1.58 2006/03/23 18:23:32 roberto Exp $
** Lexical Analyzer
** See Copyright Notice in lua.h
*/

#ifndef llex_h
#define llex_h

#include "lobject.h"
#include "lzio.h"

#define FIRST_RESERVED	257

#define TOKEN_LEN	(sizeof("function")/sizeof(char))

enum RESERVED {

  TK_AND = FIRST_RESERVED, TK_BREAK,
  TK_DO, TK_ELSE, TK_ELSEIF, TK_END, TK_FALSE, TK_FOR, TK_FUNCTION,
  TK_IF, TK_IN, TK_LOCAL, TK_NIL, TK_NOT, TK_OR, TK_REPEAT,
  TK_RETURN, TK_THEN, TK_TRUE, TK_UNTIL, TK_WHILE,

  TK_CONCAT, TK_DOTS, TK_EQ, TK_GE, TK_LE, TK_NE, TK_NUMBER,
  TK_NAME, TK_STRING, TK_EOS
};

#define NUM_RESERVED	(cast(int, TK_WHILE-FIRST_RESERVED+1))

LUAI_DATA const char *const luaX_tokens [];

typedef union {
  lua_Number r;
  TString *ts;
} SemInfo;  

typedef struct Token {
  int token;
  SemInfo seminfo;
} Token;

typedef struct LexState {
  int current;  
  int linenumber;  
  int lastline;  
  Token t;  
  Token lookahead;  
  struct FuncState *fs;  
  struct lua_State *L;
  ZIO *z;  
  Mbuffer *buff;  
  TString *source;  
  char decpoint;  
} LexState;

LUAI_FUNC void luaX_init (lua_State *L);
LUAI_FUNC void luaX_setinput (lua_State *L, LexState *ls, ZIO *z,
                              TString *source);
LUAI_FUNC TString *luaX_newstring (LexState *ls, const char *str, size_t l);
LUAI_FUNC void luaX_next (LexState *ls);
LUAI_FUNC void luaX_lookahead (LexState *ls);
LUAI_FUNC void luaX_lexerror (LexState *ls, const char *msg, int token);
LUAI_FUNC void luaX_syntaxerror (LexState *ls, const char *s);
LUAI_FUNC const char *luaX_token2str (LexState *ls, int token);

#endif
