// ############################################################################################
//  tut01d.cc
//  A tutorial application for VortexGE
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// The VortexGE include file
#include <VMain.h>


// Specialized basic application class
//
// NOTE: * We derive it from CBasicApplication which by default supports FPS control
//       * If FPS control is not needed, it must be changed to :
//           class MyApplication : public CBasicApplication<CEventHandler> {
//             ...
//             ...
//             ...
//           };
class MyApplication : public CBasicApplication<> {
  public:
    // Here we invoke the CBasicApplication constructor so that
    // we can define our window title
    inline MyApplication()
    : CBasicApplication<>("Tutorial 1d : Using the CBasicApplication Class")
    {}
    // Reimplement only the needed event handlers
    virtual void EventButtonRelease(EBtnSymbol eButtonSym) {
      if(eButtonSym == VBS_MB1) EventCloseWindow();
    }
    virtual void EventUpdateFrame(bool bWindowIsActive);
};


// Main program
int main()
{
  // Create the application class
  MyApplication csMyApp;
  csMyApp.iVideoEnv().SetForegroundColor(csMyApp.iVideoEnv().CreateColor(255, 192, 255));

  // Execute the application
  if(csMyApp.DoMainLoop() < 0) return(-1);
  csMyApp.iVideoEnv().MessageBox("End Program", "Click \"OK\" to exit ...");
  return(0);
}


// MyApplication implementation
void MyApplication::EventUpdateFrame(bool bWindowIsActive)
{
  // Just exit if the window is inactive
  if(!bWindowIsActive) return;

  // Update audio
  iAudioEnv().UpdateBuffer();

  // Update video
  i3DEnv().ClearRenderBuffer(V_RGB(12, 12, 16));
  i3DEnv().ClearZBuffer();
  iVideoEnv().PutRenderImageOnDisplay();
  iVideoEnv().DrawText(10, 20, GetFPSInfoString());
  iVideoEnv().DrawText(10, 50, "Click the left mouse button to exit ...");
  iVideoEnv().UpdateDisplay();
}

