// ############################################################################################
//  PixelColorInterpolatorFractals.cc
//  Implementation of various CPCIFractal derivative classes
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "PixelColorInterpolatorFractals.h"


namespace vge {

//
// Target Box
//
void FractalTargetBox::DrawSeed(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  // Convert the n32Y1 and n32Y2 from standard screen coordinate to cartesian space
  n32Y1 = _pcsIBuild->GetHeight() - n32Y1;
  n32Y2 = _pcsIBuild->GetHeight() - n32Y2;

  // Draw the seed
  PutPixel(n32X1, n32Y1, 0, MAX_COLOR_VALUE, MAX_COLOR_VALUE / 4);
  PutPixel(n32X2, n32Y1, 0, MAX_COLOR_VALUE, MAX_COLOR_VALUE / 4);
  PutPixel(n32X1, n32Y2, 0, MAX_COLOR_VALUE, MAX_COLOR_VALUE / 4);
  PutPixel(n32X2, n32Y2, 0, MAX_COLOR_VALUE, MAX_COLOR_VALUE / 4);
}


//
// Nebula
//
void FractalNebula::DrawSeed(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  // Convert the n32Y1 and n32Y2 from standard screen coordinate to cartesian space
  n32Y1 = _pcsIBuild->GetHeight() - n32Y1;
  n32Y2 = _pcsIBuild->GetHeight() - n32Y2;

  // Draw the seed
  PutPixel(n32X1, n32Y1, MAX_COLOR_VALUE, 0, 0);
  PutPixel(n32X1, n32Y2, 0, 0, MAX_COLOR_VALUE);
  PutPixel(n32X2, n32Y2, 0, MAX_COLOR_VALUE, 0);
}

uint8_t FractalNebula::RandomFactor(int32_t n32Distance)
{
  return(((n32Distance < 0) ? (-n32Distance) : n32Distance) / 8);
}


//
// Coral
//
void FractalCoral::DrawSeed(int32_t n32X1, int32_t n32Y1, int32_t n32X2, int32_t n32Y2)
{
  // Convert the n32Y1 and n32Y2 from standard screen coordinate to cartesian space
  n32Y1 = _pcsIBuild->GetHeight() - n32Y1;
  n32Y2 = _pcsIBuild->GetHeight() - n32Y2;

  // Draw the seed
  PutPixel(n32X1, n32Y1, MAX_COLOR_VALUE, 0, 0);
  PutPixel(n32X2, n32Y1, 0, MAX_COLOR_VALUE, MAX_COLOR_VALUE);
  PutPixel(n32X1, n32Y2, 0, 0, MAX_COLOR_VALUE);
  PutPixel(n32X2, n32Y2, 0, MAX_COLOR_VALUE, 0);
}

uint8_t FractalCoral::RandomFactor(int32_t n32Distance)
{
  UNUSED_PARAM(n32Distance);
  return(6);
}

} // namespace vge
