// ############################################################################################
//  Environment.cc
//  Implementation of CAudioEnv class using dummy (NULL) audio system
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// VortexGE headers
#include "ADMYEnvironment.h"
#include "ADMYSoundBuffer.h"


namespace vge {

//
// The dummy audio environment implementation
//
VCLASS_SINGLETON_IMPLEMENT(CDummyAudioEnv)

CDummyAudioEnv::CDummyAudioEnv()
{
}

CDummyAudioEnv::~CDummyAudioEnv() throw()
{
}

void CDummyAudioEnv::DoQuickInit(const char *pcDeviceName)
{
  UNUSED_PARAM(pcDeviceName);
}

// NOTE: If pcResult is NULL, this function will only check the
//       availability of the device.
size_t CDummyAudioEnv::QueryDeviceName(char *pcResult, size_t bufferSize)
{
  // Just return "dummy"
  if(pcResult && bufferSize) safe_strcpy(pcResult, "dummy", bufferSize);
  return(5 + 1); // strlen("dummy") == 5
}

void CDummyAudioEnv::OpenDevice(const char *pcDeviceName)
{
  UNUSED_PARAM(pcDeviceName);
}

void CDummyAudioEnv::InitializeDevice(const TAudioSettings *ptPrefered, TAudioSettings *ptRealized)
{
  if((ptRealized) && (ptRealized)) memcpy(ptRealized, ptPrefered, sizeof(TAudioSettings));
  VWarning(_NFL_, "Using dummy (NULL) audio system !!!");
}

void CDummyAudioEnv::GetAudioSpec(uint32_t *un32Channels, uint32_t *un32Bits, uint32_t *un32Rates, ESoundBufferFormat *eFormat, uint32_t *un32FragSize) const
{
  *un32Channels = 0;
  *un32Bits     = 0;
  *un32Rates    = 0;
  *eFormat      = ESB_Undefined;
  *un32FragSize = 0;
}

void CDummyAudioEnv::SetMidiPlayer(CMidiPlayer *pcsPlayer)
{
  UNUSED_PARAM(pcsPlayer);
}

CSoundBuffer *CDummyAudioEnv::NewSoundBuffer()
{
  return(_nothrow_new CDummySoundBuffer);
}

} // namespace vge
