/** \file VX11Dialog3.h */
// ############################################################################################
//  VX11Dialog3.h
//  Definition & implementation of dialog classes used by CVideoEnv class
//  using X Window system (Xlib)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This library is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// NOTE: This file is meant to be included from within
//       "lib/VideoEnvironment/VID_X11/VX11Environment.h"


#ifndef _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG3_H_
#define _VIDEO_ENVIRONMENT___VID_X11___VX11_DIALOG3_H_


namespace vge {

  //!
  //! \brief List-box control
  //!
  //! \api beta
  class VGE_EXPORT CListBox : public _CControl {
    public:
      static const uint32_t INVALID_INDEX = 0xFFFFFFFF;

    public:
      CListBox();
      virtual ~CListBox() throw();

      // Initialization functions
      //!
      //! Initialize control using the given parameters.
      //! \param[in] vidEnv               The video environment in which this control will be drawn.
      //! \param[in] pcsParent            Parent control of this control.
      //! \param[in] n32X,n32Y            Starting position of this control.
      //! \param[in] un32Width,un32Height Size of this control.
      void InitControl(CVideoEnv &vidEnv, const _CControl *pcsParent,
                       int32_t n32X, int32_t n32Y, 
                       uint32_t un32Width, uint32_t un32Height);

      // Control functions
      //!
      //! Add a new item.
      void AddItem(const string &newItem);
      //!
      //! Add new items.
      void AddItems(const vector<string> &newItems);
      //!
      //! Add new items.
      void AddItems(uint32_t un32ItemCount, const char *pcItem, ...);
      //!
      //! Delete an item.
      void DelItem(uint32_t un32ItemIndex);
      //!
      //! Delete an item.
      void DelItem(const string &item, bool bCaseSensitive = true);
      //!
      //! Delete items.
      void DelItems(uint32_t un32IndexCount, uint32_t un32ItemIndex, ...);
      //!
      //! Delete items.
      void DelItems(uint32_t un32IndexCount, const uint32_t *pun32ItemIndexes);
      //!
      //! Delete items.
      void DelItems(const vector<string> &items, bool bCaseSensitive = true);
      //!
      //! Delete all items.
      void DelAllItems();
      //!
      //! Get the index of the selected item.
      uint32_t GetSelectedIndex() const;
      //!
      //! Get the selected item.
      const char *GetSelectedItem() const;
      //!
      //! Select an item with the given index.
      void SelectItem(uint32_t un32ItemIndex);
      //!
      //! Select an item with the given name.
      void SelectItem(const string &item, bool bCaseSensitive = true);
      //!
      //! Get an array to all items.
      inline vector<string> &GetAllItems()
      { return(_sItems); }
      //!
      //! Get an array to all items.
      inline const vector<string> &GetAllItems() const
      { return(_sItems); }
      //!
      //! Find an item with the given name.
      bool FindItem(uint32_t *pun32RetIndex, const string &item, bool bCaseSensitive = true) const;
      //!
      //! Find an item with the given name.
      inline uint32_t FindItem(const string &item, bool bCaseSensitive = true) const;
      //!
      //! Check if an item with the given name is exist or not.
      inline bool ItemIsExist(const string &item, bool bCaseSensitive = true) const;
      //!
      //! Get the number of total items stored.
      inline uint32_t GetTotalItemCount() const { return(_un32Total); }
      //!
      //! If the mouse pointer is in the control's area, toggle the selection.
      bool ToggleValue(int32_t n32MouseX, int32_t n32MouseY);

      // Drawing functions
      virtual void DrawControl() const;
      virtual void DrawControl(int32_t n32MouseX, int32_t n32MouseY, bool bPressed = false) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bUpPressed, bool bDnPressed) const;
      //!
      //! Draw the control in its normal or pressed appearance.
      virtual void DrawControl(bool bUpPressed, bool bDnPressed, bool bSlPressed, uint32_t un32TxPressed) const;

      // For backward compatibility with previous versions
      //!
      //! \deprecated For backward compatibility with previous versions, will be removed soon.
      inline void InitControl(CVideoEnv *pcsVidEnv, const _CControl *pcsParent,
                              int32_t n32X, int32_t n32Y, uint32_t un32Width, uint32_t un32Height)
      { InitControl(*pcsVidEnv, pcsParent, n32X, n32Y, un32Width, un32Height); }

#define _vmcm mutable
    protected:
      XRectangle     _tOutline;     // Outline
      XRectangle     _tMain;        // Main text area
      XRectangle     _tSLB;         // Slider and button area
_vmcm XRectangle     _tSelItem;     // Selected item area
      XSegment       _atMain[4];    // Main segments
      XSegment       _tUp, _tDn;    // Buttons' area
      XSegment       _atUp[4];      // Up button segments
      XSegment       _atDn[4];      // Down button segments
      XSegment       _atArrow[8];   // Arrows segments
      XSegment       _tSlider;      // Slider's area
_vmcm XSegment       _atSlider[13]; // Slider segments
_vmcm bool           _bUseSlider;   // Use slider flag
      uint32_t       _un32AH;       // Area height
_vmcm uint32_t       _un32FH;       // Font height
_vmcm bool           _bRecalcFH;    // Flag to recalculate the font height
      uint32_t       _un32Yt;       // Start Y coordinate of text
      vector<string> _sItems;       // Item array
      uint32_t       _un32CurSel;   // Currently selected item
      uint32_t       _un32Total;    // Total items
      uint32_t       _un32StartDsp; // Displayed item start index
      uint32_t       _un32MaxDsp;   // Displayed item maximum start index
_vmcm uint32_t       _un32DspMax;   // Total number of item can be displayed
#undef _vmcm

      // Friend classes
      friend class CVideoEnv;
  };

  uint32_t CListBox::FindItem(const string &item, bool bCaseSensitive) const
  {
    uint32_t tmp;
    if(FindItem(&tmp, item, bCaseSensitive)) return(tmp);
    return(INVALID_INDEX);
  }

  bool CListBox::ItemIsExist(const string &item, bool bCaseSensitive) const
  {
    uint32_t tmp;
    return(FindItem(&tmp, item, bCaseSensitive));
  }


  //
  // Drop-down-list control
  //
  //       < NOT IMPLEMENTED YET !!!>


  //
  // Pop-up-menu control
  //
  //       < NOT IMPLEMENTED YET !!!>


  //
  // Menu-bar control
  //
  //       < NOT IMPLEMENTED YET !!!>

} // namespace vge
#endif
