// ############################################################################################
//  bench1.cc
//  Benchmark memory transfer operations
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// Include the benchmarking application header file
#include "bench.h"


//
// Helper function
//
inline void ResetMMX()
{
  asm
  (
#ifdef _ASM_3DNOW_
    "femms"
#else
    "emms"
#endif
    ::: "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm6", "mm7",
        "st", "st(1)", "st(2)", "st(3)", "st(4)", "st(5)", "st(6)", "st(7)"
  );
}


//
// Functions to be benchmarked
//

// Zero memory
inline void ZeroMem4X8(uint32_t *buff, size_t bufferSizeDividedBy32)
{
  uint32_t *tmp = buff;
  for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
    *tmp++ = 0; *tmp++ = 0; *tmp++ = 0; *tmp++ = 0;
    *tmp++ = 0; *tmp++ = 0; *tmp++ = 0; *tmp++ = 0;
  }
}

inline void ZeroMem8X4(uint32_t *buff, size_t bufferSizeDividedBy32)
{
  uint64_t *tmp = reinterpret_cast<uint64_t*>(buff);
  for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
    *tmp++ = 0; *tmp++ = 0; *tmp++ = 0; *tmp++ = 0;
  }
}

#ifdef _ASM_ASM_
inline void ZeroMem4X8ASM(uint32_t *buff, size_t bufferSizeDividedBy32)
{
#ifdef _ASM_A64_
  bufferSizeDividedBy32 *= 4;
#else
  bufferSizeDividedBy32 *= 8;
#endif
  asm volatile
  (
    "cld       \n\t"
#ifdef _ASM_A64_
    "rep stosq     "
#else
    "rep stosl     "
#endif
    : "=&D"(buff), "=c"(bufferSizeDividedBy32)
    :   "0"(buff),  "1"(bufferSizeDividedBy32), "a"(0)
    : "memory"
  );
}
#endif

#ifdef _ASM_MMX_
inline void ZeroMem8X4MMX(uint32_t *buff, size_t bufferSizeDividedBy32)
{
  asm ( "pxor %%mm0, %%mm0" ::: "mm0", "st" );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "movq %%mm0, %0 \n\t"
      "movq %%mm0, %1 \n\t"
      "movq %%mm0, %2 \n\t"
      "movq %%mm0, %3     "
      : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
    );
    buff += 8;
  }
  ResetMMX();
}
#endif

#ifdef _ASM_SSE_
inline void ZeroMem8X4SSE(uint32_t *buff, size_t bufferSizeDividedBy32)
{
  asm ( "pxor %%mm0, %%mm0" ::: "mm0", "st" );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "movntq %%mm0, %0 \n\t"
      "movntq %%mm0, %1 \n\t"
      "movntq %%mm0, %2 \n\t"
      "movntq %%mm0, %3     "
      : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
    );
    buff += 8;
  }
  ResetMMX();
}
#endif

#ifdef _ASM_SSE_
inline void ZeroMem16X2SSE(uint32_t *buff, size_t bufferSizeDividedBy32)
{
  asm ( "xorps %%xmm0, %%xmm0" ::: "xmm0" );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "movntps %%xmm0, %0 \n\t"
      "movntps %%xmm0, %1     "
      : "=m"(buff[0]), "=m"(buff[4])
    );
    buff += 8;
  }
}
#endif

// Fill memory
inline void FillMem4X8(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
{
  uint32_t *tmp = buff;
  for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
    *tmp++ = value; *tmp++ = value; *tmp++ = value; *tmp++ = value;
    *tmp++ = value; *tmp++ = value; *tmp++ = value; *tmp++ = value;
  }
}

inline void FillMem8X4(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
{
  const uint64_t  val = (uint64_t(value) << 32) | uint64_t(value);
        uint64_t *tmp = reinterpret_cast<uint64_t*>(buff);
  for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
    *tmp++ = val; *tmp++ = val; *tmp++ = val; *tmp++ = val;
  }
}

#ifdef _ASM_ASM_
inline void FillMem4X8ASM(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
{
#ifdef _ASM_A64_
  const uint64_t value64 = (uint64_t(value) << 32) | uint64_t(value);
  bufferSizeDividedBy32 *= 4;
#else
  bufferSizeDividedBy32 *= 8;
#endif
  asm volatile
  (
    "cld       \n\t"
#ifdef _ASM_A64_
    "rep stosq     "
#else
    "rep stosl     "
#endif
    : "=&D"(buff), "=c"(bufferSizeDividedBy32)
#ifdef _ASM_A64_
    :   "0"(buff),  "1"(bufferSizeDividedBy32), "a"(value64)
#else
    :   "0"(buff),  "1"(bufferSizeDividedBy32), "a"(value)
#endif
    : "memory"
  );
}
#endif

#ifdef _ASM_MMX_
inline void FillMem8X4MMX(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
{
  asm
  (
    "movd         %0, %%mm0 \n\t"
    "punpckldq %%mm0, %%mm0     "
     :: "m"(value)
      : "mm0", "st"
  );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "movq %%mm0, %0\n\t"
      "movq %%mm0, %1\n\t"
      "movq %%mm0, %2\n\t"
      "movq %%mm0, %3    "
      : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
    );
    buff += 8;
  }
  ResetMMX();
}
#endif

#ifdef _ASM_SSE_
inline void FillMem8X4SSE(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
{
  asm
  (
    "movd         %0, %%mm0 \n\t"
    "punpckldq %%mm0, %%mm0     "
     :: "m"(value)
      : "mm0", "st"
  );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "movntq %%mm0, %0\n\t"
      "movntq %%mm0, %1\n\t"
      "movntq %%mm0, %2\n\t"
      "movntq %%mm0, %3    "
      : "=m"(buff[0]), "=m"(buff[2]), "=m"(buff[4]), "=m"(buff[6])
    );
    buff += 8;
  }
  ResetMMX();
}
#endif

#ifdef _ASM_SSE_
inline void FillMem16X2SSE(uint32_t *buff, uint32_t value, size_t bufferSizeDividedBy32)
{
  asm
  (
    "movd           %0, %%xmm0 \n\t"
    "shufps $0, %%xmm0, %%xmm0     "
     :: "m"(value)
      : "xmm0"
  );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "movntps %%xmm0, %0\n\t"
      "movntps %%xmm0, %1    "
      : "=m"(buff[0]), "=m"(buff[4])
    );
    buff += 8;
  }
}
#endif

// Copy memory
inline void CopyMem4X8(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
{
        uint32_t *d = dstBuff;
  const uint32_t *s = srcBuff;
  for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
    *d++ = *s++; *d++ = *s++; *d++ = *s++; *d++ = *s++;
    *d++ = *s++; *d++ = *s++; *d++ = *s++; *d++ = *s++;
  }
}

inline void CopyMem8X4(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
{
        uint64_t *d = reinterpret_cast<      uint64_t*>(dstBuff);
  const uint64_t *s = reinterpret_cast<const uint64_t*>(srcBuff);
  for(size_t i = 0; i < bufferSizeDividedBy32; ++i) {
    *d++ = *s++; *d++ = *s++; *d++ = *s++; *d++ = *s++;
  }
}

#ifdef _ASM_ASM_
inline void CopyMem4X8ASM(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
{
#ifdef _ASM_A64_
  bufferSizeDividedBy32 *= 4;
#else
  bufferSizeDividedBy32 *= 8;
#endif
  asm volatile
  (
    "cld        \n\t"
#ifdef _ASM_A64_
    "rep movsq      "
#else
    "rep movsl      "
#endif
    : "=&D"(dstBuff), "=&S"(srcBuff), "=c"(bufferSizeDividedBy32) 
    :   "0"(dstBuff),   "1"(srcBuff),  "2"(bufferSizeDividedBy32)
    : "memory"
  );
}
#endif

#ifdef _ASM_MMX_
inline void CopyMem8X4MMX(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
{
  while(bufferSizeDividedBy32--) {
    asm
    (
      "movq    %4, %%mm0 \n\t"
      "movq    %5, %%mm1 \n\t"
      "movq    %6, %%mm2 \n\t"
      "movq    %7, %%mm3 \n\t"
      "movq %%mm0, %0    \n\t"
      "movq %%mm1, %1    \n\t"
      "movq %%mm2, %2    \n\t"
      "movq %%mm3, %3        "
      : "=m"(dstBuff[0]), "=m"(dstBuff[2]), "=m"(dstBuff[4]), "=m"(dstBuff[6])
      :  "m"(srcBuff[0]),  "m"(srcBuff[2]),  "m"(srcBuff[4]),  "m"(srcBuff[6])
      : "mm0", "mm1", "mm2", "mm3", "st", "st(1)", "st(2)", "st(3)"
    );
    dstBuff += 8;
    srcBuff += 8;
  }
  ResetMMX();
}
#endif

#ifdef _ASM_SSE_
inline void CopyMem8X4SSE(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
{
  asm
  (
    "prefetchnta %0 \n\t"
    "prefetchnta %1 \n\t"
    "prefetchnta %2     "
    :: "m"(srcBuff[0]), "m"(srcBuff[8]), "m"(srcBuff[16])
  );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "prefetchnta %8        \n\t"
      "movq        %4, %%mm0 \n\t"
      "movq        %5, %%mm1 \n\t"
      "movq        %6, %%mm2 \n\t"
      "movq        %7, %%mm3 \n\t"
      "movntq   %%mm0, %0    \n\t"
      "movntq   %%mm1, %1    \n\t"
      "movntq   %%mm2, %2    \n\t"
      "movntq   %%mm3, %3        "
      : "=m"(dstBuff[0]), "=m"(dstBuff[2]), "=m"(dstBuff[4]), "=m"(dstBuff[6])
      :  "m"(srcBuff[0]),  "m"(srcBuff[2]),  "m"(srcBuff[4]),  "m"(srcBuff[6]), "m"(srcBuff[24])
      : "mm0", "mm1", "mm2", "mm3", "st", "st(1)", "st(2)", "st(3)"
    );
    dstBuff += 8;
    srcBuff += 8;
  }
  ResetMMX();
}
#endif

#ifdef _ASM_SSE_
inline void CopyMem16X2SSE(uint32_t *dstBuff, const uint32_t *srcBuff, size_t bufferSizeDividedBy32)
{
  asm
  (
    "prefetchnta %0 \n\t"
    "prefetchnta %1 \n\t"
    "prefetchnta %2     "
    :: "m"(srcBuff[0]), "m"(srcBuff[8]), "m"(srcBuff[16])
  );
  while(bufferSizeDividedBy32--) {
    asm
    (
      "prefetchnta %4         \n\t"
      "movaps      %2, %%xmm0 \n\t"
      "movaps      %3, %%xmm1 \n\t"
      "movntps %%xmm0, %0     \n\t"
      "movntps %%xmm1, %1         "
      : "=m"(dstBuff[0]), "=m"(dstBuff[4])
      :  "m"(srcBuff[0]),  "m"(srcBuff[4]), "m"(srcBuff[24])
      : "xmm0", "xmm1"
    );
    dstBuff += 8;
    srcBuff += 8;
  }
}
#endif


//
// Helper functions
//

// Get the system ticks
const uint64_t TICKS_PER_SEC = uint64_t(1000000);
inline uint64_t GetTicks()
{
  struct timeval tv;
  gettimeofday(&tv, 0);
  return(uint64_t(tv.tv_sec) * TICKS_PER_SEC + tv.tv_usec);
}

// Logger
void Log(const char *pcMsg, ...)
{
  // Temporary variables
  char    acBuff[1024];
  va_list vlTmp;

  // Extract the pcMsg
  va_start(vlTmp, pcMsg);
    vsnprintf(acBuff, sizeof(acBuff), pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
  printf(acBuff);
}

// Aligned memory allocation
void *amalloc(uint32_t un32Size, uint32_t un32Alignment = 16)
{
  // Temporary variables
  void    *pvTmp;
  uint32_t un32R;

  // Make sure that the alignment >= sizeof(void *)
  if(un32Alignment < sizeof(void *)) un32Alignment = sizeof(void *);

  // Allocate aligned memory
  un32R = posix_memalign(&pvTmp, un32Alignment, un32Size);
  if(un32R) {
    if(un32R == EINVAL) printf("The alignment must be a power of two and a multiple of %zd (size of void *)\n", sizeof(void *));
    if(un32R == ENOMEM) printf("Not enough memory\n");
    return(NULL);
  }
  return(pvTmp);
}

// Validation functions
void vfill(u_char *pucBuffer, uint32_t un32Size, u_char ucVal)
{
  memset(pucBuffer, ucVal, un32Size);
}

bool vcheck(u_char *pucBuffer, uint32_t un32Size, u_char ucVal)
{
  u_char  *pucBuff = pucBuffer;
  uint32_t un32Cnt;
  for(un32Cnt = 0; un32Cnt < un32Size; un32Cnt++)
    if(*pucBuff++ != ucVal) return(false);
  return(true);
}


//
// Main program (benchmaker)
//
#define TOT_ITER 256

int main()
{
  // Allocate dummy memory blocks
  const uint32_t  un32Size   = 4 * 1024 * 1024; 
  const uint32_t  un32Size32 = un32Size / 32;
  const double    facTicks   = 1.0 / double(TOT_ITER);
  const double    facSpeed   = double(TICKS_PER_SEC) * double(TOT_ITER) * (double(un32Size) / double(1024*1024));
  u_char         *pucSrc     = (u_char*) ::amalloc(un32Size);
  u_char         *pucDst     = (u_char*) ::amalloc(un32Size);

  // For benchmarking
  uint64_t un64TS, un64TE, un64TC;
  uint32_t un32Cnt;
  bool     bVal;

  // Benchmark ZeroMemory functions
  Log("\nBenchmarking ZeroMemory functions ...\n");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    memset(pucDst, 0, un32Size);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); memset(pucDst, 0, un32Size); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  C++ memset     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ::ZeroMem4X8((uint32_t*) pucDst, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ::ZeroMem4X8((uint32_t*) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem4X8     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ZeroMem8X4((uint32_t*) pucDst, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem8X4((uint32_t*) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem8X4     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#ifdef _ASM_ASM_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ZeroMem4X8ASM((uint32_t*) pucDst, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem4X8ASM((uint32_t*) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem4X8ASM  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  ZeroMem4X8ASM  : <<< ASM is not enabled >>>\n");
#endif
#ifdef _ASM_MMX_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ZeroMem8X4MMX((uint32_t*) pucDst, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem8X4MMX((uint32_t*) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem8X4MMX  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  ZeroMem8X4MMX  : <<< MMX is not enabled >>>\n");
#endif
#ifdef _ASM_SSE_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ZeroMem8X4SSE((uint32_t*) pucDst, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem8X4SSE((uint32_t*) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem8X4SSE  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ZeroMem16X2SSE((uint32_t*) pucDst, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  vfill(pucDst, un32Size, 0xFF); ZeroMem16X2SSE((uint32_t*) pucDst, un32Size32); bVal = vcheck(pucDst, un32Size, 0x00);
  Log("  ZeroMem16X2SSE : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  ZeroMem8X4SSE  : <<< SSE is not enabled >>>\n");
  Log("  ZeroMem16X2SSE : <<< SSE is not enabled >>>\n");
#endif

  // Benchmark FillMemory functions
  Log("\nBenchmarking FillMemory functions ...\n");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    memset(pucDst, 0xF0F0F0F0, un32Size);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); memset(pucDst, 0xF0, un32Size); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  C++ memset     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ::FillMem4X8((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); ::FillMem4X8((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem4X8     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    FillMem8X4((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem8X4((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem8X4     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#ifdef _ASM_ASM_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    FillMem4X8ASM((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem4X8ASM((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem4X8ASM  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  FillMem4X8ASM  : <<< ASM is not enabled >>>\n");
#endif
#ifdef _ASM_MMX_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    FillMem8X4MMX((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem8X4MMX((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem8X4MMX  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  FillMem8X4MMX  : <<< MMX is not enabled >>>\n");
#endif
#ifdef _ASM_SSE_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    FillMem8X4SSE((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem8X4SSE((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem8X4SSE  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    FillMem16X2SSE((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); FillMem16X2SSE((uint32_t*) pucDst, 0xF0F0F0F0, un32Size32); bVal = vcheck(pucDst, un32Size, 0xF0);
  Log("  FillMem16X2SSE : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  FillMem8X4SSE  : <<< SSE is not enabled >>>\n");
  Log("  FillMem16X2SSE : <<< SSE is not enabled >>>\n");
#endif

  // Benchmark CopyMemory functions
  memset(pucSrc, 0x88, un32Size);
  Log("\nBenchmarking CopyMemory functions ...\n");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    memcpy(pucDst, pucSrc, un32Size);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); memcpy(pucDst, pucSrc, un32Size); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  C++ memcpy     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    ::CopyMem4X8((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); ::CopyMem4X8((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem4X8     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    CopyMem8X4((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem8X4((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem8X4     : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#ifdef _ASM_ASM_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    CopyMem4X8ASM((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem4X8ASM((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem4X8ASM  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  CopyMem4X8ASM  : <<< ASM is not enabled >>>\n");
#endif
#ifdef _ASM_MMX_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    CopyMem8X4MMX((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem8X4MMX((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem8X4MMX  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  CopyMem8X4MMX  : <<< MMX is not enabled >>>\n");
#endif
#ifdef _ASM_SSE_
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    CopyMem8X4SSE((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem8X4SSE((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem8X4SSE  : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
  for(un64TC = 0, un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) {
    un64TS = GetTicks();
    CopyMem16X2SSE((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32);
    un64TE = GetTicks();
    un64TC += (un64TE - un64TS);
  }
  memset(pucDst, 0, un32Size); CopyMem16X2SSE((uint32_t*) pucDst, (uint32_t*) pucSrc, un32Size32); bVal = vcheck(pucDst, un32Size, 0x88);
  Log("  CopyMem16X2SSE : ");
  Log("%7.2f ticks (%7.2f MB/s) : %s\n", facTicks*double(un64TC), facSpeed/double(un64TC), bVal?"OK":"BAD");
#else
  Log("  CopyMem8X4SSE  : <<< SSE is not enabled >>>\n");
  Log("  CopyMem16X2SSE : <<< SSE is not enabled >>>\n");
#endif
  Log("\n");

  // Free memories
  free(pucSrc);
  free(pucDst);

  // End
  return(0);
}
