// ############################################################################################
//  bench2.cc
//  Benchmark matrix operations
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// Include the benchmarking application header file
#include "bench.h"


//
// Functions to be benchmarked
//

// Matrix copy
inline void MCopy(float aafDstMat[4][4], const float aafSrcMat[4][4])
{
  aafDstMat[0][0] = aafSrcMat[0][0]; aafDstMat[0][1] = aafSrcMat[0][1]; aafDstMat[0][2] = aafSrcMat[0][2]; aafDstMat[0][3] = aafSrcMat[0][3];
  aafDstMat[1][0] = aafSrcMat[1][0]; aafDstMat[1][1] = aafSrcMat[1][1]; aafDstMat[1][2] = aafSrcMat[1][2]; aafDstMat[1][3] = aafSrcMat[1][3];
  aafDstMat[2][0] = aafSrcMat[2][0]; aafDstMat[2][1] = aafSrcMat[2][1]; aafDstMat[2][2] = aafSrcMat[2][2]; aafDstMat[2][3] = aafSrcMat[2][3];
  aafDstMat[3][0] = aafSrcMat[3][0]; aafDstMat[3][1] = aafSrcMat[3][1]; aafDstMat[3][2] = aafSrcMat[3][2]; aafDstMat[3][3] = aafSrcMat[3][3];
}

#ifdef _ASM_MMX_
inline void MCopyMMX(float aafDstMat[4][4], const float aafSrcMat[4][4])
{
  uint32_t edi;
  asm volatile
  (
    "movq   (%%esi), %%mm0 \n"
    "movq  8(%%esi), %%mm1 \n"
    "movq 16(%%esi), %%mm2 \n"
    "movq 24(%%esi), %%mm3 \n"
    "movq 32(%%esi), %%mm4 \n"
    "movq 40(%%esi), %%mm5 \n"
    "movq 48(%%esi), %%mm6 \n"
    "movq 56(%%esi), %%mm7 \n"
    "movq %%mm0,   (%%edi) \n"
    "movq %%mm1,  8(%%edi) \n"
    "movq %%mm2, 16(%%edi) \n"
    "movq %%mm3, 24(%%edi) \n"
    "movq %%mm4, 32(%%edi) \n"
    "movq %%mm5, 40(%%edi) \n"
    "movq %%mm6, 48(%%edi) \n"
    "movq %%mm7, 56(%%edi) \n"
    "emms                  \n"
    : "=&D" (edi)                        \
    :   "0" (aafDstMat), "S" (aafSrcMat) \
    : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm6", "mm7", "memory"
  );
}
#endif

#ifdef _ASM_SSE_
inline void MCopySSE(float aafDstMat[4][4], const float aafSrcMat[4][4])
{
  uint32_t edi;
  asm volatile
  (
    "movaps   (%%esi), %%xmm0 \n"
    "movaps 16(%%esi), %%xmm1 \n"
    "movaps 32(%%esi), %%xmm2 \n"
    "movaps 48(%%esi), %%xmm3 \n"
    "movaps %%xmm0,   (%%edi) \n"
    "movaps %%xmm1, 16(%%edi) \n"
    "movaps %%xmm2, 32(%%edi) \n"
    "movaps %%xmm3, 48(%%edi) \n"
    : "=&D" (edi)                        \
    :   "0" (aafDstMat), "S" (aafSrcMat) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "memory"
  );
}
#endif

// Matrix indentity
inline void M1(float aafDstMatrix[4][4])
{
  aafDstMatrix[0][0] = 1; aafDstMatrix[0][1] = 0; aafDstMatrix[0][2] = 0; aafDstMatrix[0][3] = 0;
  aafDstMatrix[1][0] = 0; aafDstMatrix[1][1] = 1; aafDstMatrix[1][2] = 0; aafDstMatrix[1][3] = 0;
  aafDstMatrix[2][0] = 0; aafDstMatrix[2][1] = 0; aafDstMatrix[2][2] = 1; aafDstMatrix[2][3] = 0;
  aafDstMatrix[3][0] = 0; aafDstMatrix[3][1] = 0; aafDstMatrix[3][2] = 0; aafDstMatrix[3][3] = 1;
}

#ifdef _ASM_MMX_
inline void M1MMX(float aafDstMatrix[4][4])
{
  static float _sse afOZZO[4] = {1, 0, 0, 1};
  uint32_t edi;
  asm volatile
  (
    "movq (%%esi), %%mm0   \n" //  mm0 = 1 0
    "pxor %%mm1, %%mm1     \n" //  mm1 = 0 0
    "movq 8(%%esi), %%mm2  \n" //  mm2 = 0 1
    "movq %%mm0,   (%%edi) \n" // ROW0 = 1 0 X X
    "movq %%mm1,  8(%%edi) \n" // ROW0 = 1 0 0 0
    "movq %%mm2, 16(%%edi) \n" // ROW1 = 0 1 X X
    "movq %%mm1, 24(%%edi) \n" // ROW1 = 0 1 0 0
    "movq %%mm1, 32(%%edi) \n" // ROW2 = 0 0 X X
    "movq %%mm0, 40(%%edi) \n" // ROW2 = 0 0 1 0
    "movq %%mm1, 48(%%edi) \n" // ROW3 = 0 0 X X
    "movq %%mm2, 56(%%edi) \n" // ROW3 = 0 0 0 1
    "emms                  \n"
    : "=&D" (edi)                        \
    :   "0" (aafDstMatrix), "S" (afOZZO) \
    : "mm0", "mm1", "mm2", "memory"
  );
}
#endif

#ifdef _ASM_SSE_
inline void M1SSE(float aafDstMatrix[4][4])
{
  static float fOne = float(1.0);
  uint32_t edi;
  asm volatile
    (
                                      // XPOS - 0 1 2 3
    "xorps  %%xmm0, %%xmm0        \n" // xmm0 = 0 0 0 0
    "movss (%%esi), %%xmm0        \n" // xmm0 = 1 0 0 0
    "movaps %%xmm0, %%xmm1        \n" // xmm1 = 1 0 0 0
    "movaps %%xmm0, %%xmm2        \n" // xmm2 = 1 0 0 0
    "movaps %%xmm0, %%xmm3        \n" // xmm3 = 1 0 0 0
    "shufps $0xF1, %%xmm1, %%xmm1 \n" // xmm1 = 0 1 0 0
    "shufps $0x4F, %%xmm2, %%xmm2 \n" // xmm2 = 0 0 1 0
    "shufps $0x1F, %%xmm3, %%xmm3 \n" // xmm3 = 0 0 0 1
    "movaps %%xmm0,   (%%edi)     \n" // ROW0 = 1 0 0 0
    "movaps %%xmm1, 16(%%edi)     \n" // ROW1 = 0 1 0 0
    "movaps %%xmm2, 32(%%edi)     \n" // ROW2 = 0 0 1 0
    "movaps %%xmm3, 48(%%edi)     \n" // ROW3 = 0 0 0 1
    : "=&D" (edi)                       \
    :   "0" (aafDstMatrix), "S" (&fOne) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "memory"
  );
}
#endif

// Matrix transpose
inline void MTranspose(float aafDstMat[4][4], const float aafSrcMat[4][4])
{
  aafDstMat[0][0] = aafSrcMat[0][0]; aafDstMat[0][1] = aafSrcMat[1][0]; aafDstMat[0][2] = aafSrcMat[2][0]; aafDstMat[0][3] = aafSrcMat[3][0];
  aafDstMat[1][0] = aafSrcMat[0][1]; aafDstMat[1][1] = aafSrcMat[1][1]; aafDstMat[1][2] = aafSrcMat[2][1]; aafDstMat[1][3] = aafSrcMat[3][1];
  aafDstMat[2][0] = aafSrcMat[0][2]; aafDstMat[2][1] = aafSrcMat[1][2]; aafDstMat[2][2] = aafSrcMat[2][2]; aafDstMat[2][3] = aafSrcMat[3][2];
  aafDstMat[3][0] = aafSrcMat[0][3]; aafDstMat[3][1] = aafSrcMat[1][3]; aafDstMat[3][2] = aafSrcMat[2][3]; aafDstMat[3][3] = aafSrcMat[3][3];
}

#ifdef _ASM_SSE_
inline void MTransposeSSE(float aafDstMat[4][4], const float aafSrcMat[4][4])
{
#define CHOOSE_MTransposeSSE 2

#if CHOOSE_MTransposeSSE == 0
  uint32_t esi, edi;
  asm volatile
  (
    // This is the original version :
    // Copy from buffer                    MCOL - C3 C2 C1 C0
    //                                     XPOS - F0 F1 F2 F3
    "movaps   (%%esi), %%xmm0     \n" // xmm0 = 00 01 02 03 (ROW 0)
    "movaps 16(%%esi), %%xmm1     \n" // xmm1 = 10 11 12 13 (ROW 1)
    "movaps 32(%%esi), %%xmm2     \n" // xmm2 = 20 21 22 23 (ROW 2)
    "movaps 48(%%esi), %%xmm3     \n" // xmm3 = 30 31 32 33 (ROW 3)
    // Phase 1                             XPOS - F0 F1 F2 F3
    "movaps %%xmm1, %%xmm4        \n" // xmm4 = xmm1
    "movaps %%xmm3, %%xmm5        \n" // xmm5 = xmm3
    "shufps $0x11, %%xmm0, %%xmm1 \n" // xmm1 = 11 10 01 00 (mask = 0101)
    "shufps $0xBB, %%xmm0, %%xmm4 \n" // xmm4 = 13 12 03 02 (mask = 2323)
    "shufps $0x11, %%xmm2, %%xmm3 \n" // xmm3 = 31 30 21 20 (mask = 0101)
    "shufps $0xBB, %%xmm2, %%xmm5 \n" // xmm5 = 33 32 23 22 (mask = 2323)
    // Phase 2                             XPOS - F0 F1 F2 F3
    "movaps %%xmm1, %%xmm2        \n" // xmm2 = xmm1
    "movaps %%xmm4, %%xmm0        \n" // xmm0 = xmm4
    "shufps $0x77, %%xmm3, %%xmm1 \n" // xmm1 = 00 10 20 30 (mask = 1313)
    "shufps $0x22, %%xmm3, %%xmm2 \n" // xmm2 = 01 11 21 31 (mask = 0202)
    "shufps $0x77, %%xmm5, %%xmm4 \n" // xmm4 = 02 12 22 32 (mask = 1313)
    "shufps $0x22, %%xmm5, %%xmm0 \n" // xmm0 = 03 13 23 33 (mask = 0202)
    //                                     XPOS - F0 F1 F2 F3
    // Copy to buffer                      MCOL - C3 C2 C1 C0
    "movaps %%xmm1,   (%%edi)     \n" // ROW0 = 00 10 20 30
    "movaps %%xmm2, 16(%%edi)     \n" // ROW1 = 01 11 21 31
    "movaps %%xmm4, 32(%%edi)     \n" // ROW2 = 02 12 22 32
    "movaps %%xmm0, 48(%%edi)     \n" // ROW3 = 03 13 23 33
    : "=&D" (edi),     "=&S" (esi)       \
    :   "0" (aafDstMat), "1" (aafSrcMat) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "memory"
  );
#elif CHOOSE_MTransposeSSE == 1
  uint32_t edi;
  asm volatile
  (
    // This is the optimized version (runs slightly faster than the original version) :
    //                                     MCOL - C0 C1 C2 C3
    // Copy from buffer                    XPOS - F0 F1 F2 F3
    "movaps 16(%%esi), %%xmm1     \n" // xmm1 = 10 11 12 13 (ROW 1)
    "movaps %%xmm1, %%xmm4        \n" // xmm4 = xmm1
    "movaps (%%esi), %%xmm0       \n" // xmm0 = 00 01 02 03 (ROW 0)
    "shufps $0x11, %%xmm0, %%xmm1 \n" // xmm1 = 11 10 01 00 (mask = 0101)
    "shufps $0xBB, %%xmm0, %%xmm4 \n" // xmm4 = 13 12 03 02 (mask = 2323)
    "movaps 48(%%esi), %%xmm3     \n" // xmm3 = 30 31 32 33 (ROW 3)
    "movaps %%xmm3, %%xmm5        \n" // xmm5 = xmm3
    "movaps 32(%%esi), %%xmm2     \n" // xmm2 = 20 21 22 23 (ROW 2)
    "shufps $0x11, %%xmm2, %%xmm3 \n" // xmm3 = 31 30 21 20 (mask = 0101)
    "shufps $0xBB, %%xmm2, %%xmm5 \n" // xmm5 = 33 32 23 22 (mask = 2323)
    "movaps %%xmm1, %%xmm2        \n" // xmm2 = xmm1
    "shufps $0x77, %%xmm3, %%xmm1 \n" // xmm1 = 00 10 20 30 (mask = 1313)
    "movaps %%xmm1, (%%edi)       \n" // ROW0 = 00 10 20 30
    "shufps $0x22, %%xmm3, %%xmm2 \n" // xmm2 = 01 11 21 31 (mask = 0202)
    "movaps %%xmm2, 16(%%edi)     \n" // ROW1 = 01 11 21 31
    "movaps %%xmm4, %%xmm0        \n" // xmm0 = xmm4
    "shufps $0x77, %%xmm5, %%xmm4 \n" // xmm4 = 02 12 22 32 (mask = 1313)
    "movaps %%xmm4, 32(%%edi)     \n" // ROW2 = 02 12 22 32
    "shufps $0x22, %%xmm5, %%xmm0 \n" // xmm0 = 03 13 23 33 (mask = 0202)
    "movaps %%xmm0, 48(%%edi)     \n" // ROW3 = 03 13 23 33
    : "=&D" (edi)                        \
    :   "0" (aafDstMat), "S" (aafSrcMat) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "memory"
  );
#else
    // This is the version which use GCC's built-in intrinsic functions
    typedef float v4sf __attribute__ ((vector_size(16)));

    const v4sf &sr0 = *reinterpret_cast<const v4sf*>(aafSrcMat[0]);
    const v4sf &sr1 = *reinterpret_cast<const v4sf*>(aafSrcMat[1]);
    const v4sf &sr2 = *reinterpret_cast<const v4sf*>(aafSrcMat[2]);
    const v4sf &sr3 = *reinterpret_cast<const v4sf*>(aafSrcMat[3]);

    const register v4sf tm0 = __builtin_ia32_shufps(sr1, sr0, 0x11);
    const register v4sf tm1 = __builtin_ia32_shufps(sr1, sr0, 0xBB);
    const register v4sf tm2 = __builtin_ia32_shufps(sr3, sr2, 0x11);
    const register v4sf tm3 = __builtin_ia32_shufps(sr3, sr2, 0xBB);

    *reinterpret_cast<v4sf*>(aafDstMat[0]) = __builtin_ia32_shufps(tm0, tm2, 0x77);
    *reinterpret_cast<v4sf*>(aafDstMat[1]) = __builtin_ia32_shufps(tm0, tm2, 0x22);
    *reinterpret_cast<v4sf*>(aafDstMat[2]) = __builtin_ia32_shufps(tm1, tm3, 0x77);
    *reinterpret_cast<v4sf*>(aafDstMat[3]) = __builtin_ia32_shufps(tm1, tm3, 0x22);
#endif
}
#endif

// Vector - Matrix multiplications
inline void VMMul(float afDstVec[4], const float afSrcVec[4], const float aafSrcMat[4][4])
{
  // Transform a 3D (XYZ) point/vector using a 4x4 matrix,
  //       the fourth vector component is not used (don't care)
  afDstVec[0] = (afSrcVec[0] * aafSrcMat[0][0]) +
                (afSrcVec[1] * aafSrcMat[1][0]) +
                (afSrcVec[2] * aafSrcMat[2][0]) +
                aafSrcMat[3][0];
  afDstVec[1] = (afSrcVec[0] * aafSrcMat[0][1]) +
                (afSrcVec[1] * aafSrcMat[1][1]) +
                (afSrcVec[2] * aafSrcMat[2][1]) +
                aafSrcMat[3][1];
  afDstVec[2] = (afSrcVec[0] * aafSrcMat[0][2]) +
                (afSrcVec[1] * aafSrcMat[1][2]) +
                (afSrcVec[2] * aafSrcMat[2][2]) +
                aafSrcMat[3][2];
}

#ifdef _ASM_SSE_
inline void VMMulSSE(float afDstVec[4], const float afSrcVec[4], const float aafSrcMat[4][4])
{
  // Transform a 3D (XYZ) point/vector using a 4x4 matrix,
  //       the fourth vector component is not used (don't care)
  uint32_t edi;
  asm volatile
  (
#if 0
    // This is the original version :
    //                                      XPOS - F0 F1 F2 F3
    "movss    (%%esi), %%xmm0        \n" // xmm0 = V0 XX XX XX
    "movss   4(%%esi), %%xmm1        \n" // xmm1 = V1 XX XX XX
    "movss   8(%%esi), %%xmm2        \n" // xmm2 = V2 XX XX XX
    "shufps     $0x00, %%xmm0, %%xmm0\n" // xmm0 = V0 V0 V0 V0
    "shufps     $0x00, %%xmm1, %%xmm1\n" // xmm1 = V1 V1 V1 V1
    "shufps     $0x00, %%xmm2, %%xmm2\n" // xmm2 = V2 V2 V2 V2
    "mulps    (%%ebx), %%xmm0        \n" // xmm0 = [V0 V0 V0 V0] * [00 01 02 03]
    "mulps  16(%%ebx), %%xmm1        \n" // xmm1 = [V1 V1 V1 V1] * [10 11 12 13]
    "mulps  32(%%ebx), %%xmm2        \n" // xmm2 = [V2 V2 V2 V2] * [20 21 22 23]
    "addps  48(%%ebx), %%xmm2        \n" // xmm2 = (V2 X row2) + row3
    "addps     %%xmm1, %%xmm0        \n" // xmm0 = (V0 X row0) + (V1 X row1)
    "addps     %%xmm2, %%xmm0        \n" // xmm0 = (V0 X row0) + (V1 X row1) + (V2 X row2) + row3
    "movaps    %%xmm0, (%%edi)       \n" // put the final result
#else
    // This is the optimized version (may run slightly faster than the original version) :
    //                                     XPOS - F0 F1 F2 F3
    "movss    (%%esi), %%xmm0        \n"// xmm0 = V0 XX XX XX
    "shufps     $0x00, %%xmm0, %%xmm0\n"// xmm0 = V0 V0 V0 V0
    "mulps    (%%ebx), %%xmm0        \n"// xmm0 = [V0 V0 V0 V0] * [00 01 02 03]
    "movss   4(%%esi), %%xmm1        \n"// xmm1 = V1 XX XX XX
    "shufps     $0x00, %%xmm1, %%xmm1\n"// xmm1 = V1 V1 V1 V1
    "movss   8(%%esi), %%xmm2        \n"// xmm2 = V2 XX XX XX
    "shufps     $0x00, %%xmm2, %%xmm2\n"// xmm2 = V2 V2 V2 V2
    "mulps  16(%%ebx), %%xmm1        \n"// xmm1 = [V1 V1 V1 V1] * [10 11 12 13]
    "addps     %%xmm1, %%xmm0        \n"// xmm0 = (V0 X row0) + (V1 X row1)
    "mulps  32(%%ebx), %%xmm2        \n"// xmm2 = [V2 V2 V2 V2] * [20 21 22 23]
    "addps  48(%%ebx), %%xmm2        \n"// xmm2 = (V2 X row2) + row3
    "addps     %%xmm2, %%xmm0        \n"// xmm0 = (V0 X row0) + (V1 X row1) + (V2 X row2) + row3
    "movaps    %%xmm0, (%%edi)       \n"// put the final result
#endif
    : "=&D" (edi)                                       \
    :   "0" (afDstVec), "S" (afSrcVec), "b" (aafSrcMat) \
    : "xmm0", "xmm1", "xmm2", "memory"
  );
}
#endif

#ifdef _ASM_3DNOW_
inline void VMMul3DNow(float afDstVec[4], const float afSrcVec[4], const float aafSrcMat[4][4])
{
  // Transform a 3D (XYZ) point/vector using a 4x4 matrix,
  //       the fourth vector component is not used (don't care)
  asm volatile
  (
  // These code was originally published in :
  //   Chapter 10, Optimized Matrix Multiplication
  //   AMD Athlon Processor x86 Code Optimization Guide
  //   (C) 2000 Advanced Micro Devices, Inc.
  // The GCC inline assembly form of these code was taken with some modifications from :
  //   ulaalgo.h
  //   uSTL library, ustl-0.4-3.tar.bz2
  //   http://sourceforge.net/projects/ustl
  //
    //                              XPOS -              Float #1 | Float #0
    "movq        (%0), %%mm0  \n" // mm0 =                     Y | X
    "movq       8(%0), %%mm1  \n" // mm1 =                     . | Z
    "movq       %%mm0, %%mm2  \n" // mm2 =                     Y | X
    "movq        (%1), %%mm3  \n" // mm3 =                   M01 | M00
    "punpckldq  %%mm0, %%mm0  \n" // mm0 =                     X | X
    "movq      16(%1), %%mm4  \n" // mm4 =                   M11 | M10
    "pfmul      %%mm0, %%mm3  \n" // mm3 =                 X*M01 | X*M00
    "punpckhdq  %%mm2, %%mm2  \n" // mm2 =                     Y | Y
    "pfmul      %%mm2, %%mm4  \n" // mm4 =                 Y*M11 | Y*M10
    "movq       8(%1), %%mm5  \n" // mm5 =                   M03 | M02
    "movq      24(%1), %%mm7  \n" // mm7 =                   M13 | M12
    "movq       %%mm1, %%mm6  \n" // mm6 =                     . | Z
    "pfmul      %%mm0, %%mm5  \n" // mm5 =                 X*M03 | X*M02
    "movq      32(%1), %%mm0  \n" // mm0 =                   M21 | M20
    "punpckldq  %%mm1, %%mm1  \n" // mm1 =                     Z | Z
    "pfmul      %%mm2, %%mm7  \n" // mm7 =                 Y*M13 | Y*M12
    "movq      40(%1), %%mm2  \n" // mm2 =                   M23 | M22
    "pfmul      %%mm1, %%mm0  \n" // mm0 =                 Z*M21 | Z*M20
    "pfadd      %%mm4, %%mm3  \n" // mm3 =           X*M01+Y*M11 | X*M00+Y*M10
    "movq      48(%1), %%mm4  \n" // mm4 =                   M31 | M30
    "pfmul      %%mm1, %%mm2  \n" // mm2 =                 Z*M23 | Z*M22
    "pfadd      %%mm7, %%mm5  \n" // mm5 =           X*M03+Y*M13 | X*M02+Y*M12
    "movq      56(%1), %%mm1  \n" // mm1 =                   M33 | M32
    "punpckhdq  %%mm6, %%mm6  \n" // mm6 =                     . | .
    "pfadd      %%mm0, %%mm3  \n" // mm3 =     X*M01+Y*M11+Z*M21 | X*M00+Y*M10+Z*M20
    "pfmul      %%mm6, %%mm4  \n" // mm4 =             .*M[3][1] | .*M[3][0]
    "pfmul      %%mm6, %%mm1  \n" // mm1 =             .*M[3][3] | .*M[3][2]
    "pfadd      %%mm2, %%mm5  \n" // mm5 =     X*M03+Y*M13+Z*M23 | X*M02+Y*M12+Z*M22
    "pfadd      %%mm4, %%mm3  \n" // mm3 = X*M01+Y*M11+Z*M21+M31 | X*M00+Y*M10+Z*M20+M30
    "movq       %%mm3, (%2)   \n" // store                    Y' | X'
    "pfadd      %%mm1, %%mm5  \n" // mm5 = X*M03+Y*M13+Z*M23+M33 | X*M02+Y*M12+Z*M22+M32
    "movq       %%mm5, 8(%2)  \n" // store          <don't care> | Z'
    "femms                    \n" // clear mmx state
    : : "r"(afSrcVec), "r"(aafSrcMat), "r"(afDstVec)
    : "mm0", "mm1", "mm2", "mm3", "mm4", "mm5", "mm7", "memory"
  );
}
#endif

// Matrix - Matrix multiplications
void MMMul(float aafDstMatrix[4][4], float aafSrcMatrix[4][4])
{
  float aafTmp[4][4];

  for(uint8_t un8I = 0; un8I < 4; un8I++)
    for(uint8_t un8J = 0; un8J < 4; un8J++)
      aafTmp[un8I][un8J] = (aafSrcMatrix[un8I][0] * aafDstMatrix[0][un8J]) +
        (aafSrcMatrix[un8I][1] * aafDstMatrix[1][un8J]) +
        (aafSrcMatrix[un8I][2] * aafDstMatrix[2][un8J]) +
        (aafSrcMatrix[un8I][3] * aafDstMatrix[3][un8J]);

  aafDstMatrix[0][0] = aafTmp[0][0]; aafDstMatrix[0][1] = aafTmp[0][1]; aafDstMatrix[0][2] = aafTmp[0][2]; aafDstMatrix[0][3] = aafTmp[0][3];
  aafDstMatrix[1][0] = aafTmp[1][0]; aafDstMatrix[1][1] = aafTmp[1][1]; aafDstMatrix[1][2] = aafTmp[1][2]; aafDstMatrix[1][3] = aafTmp[1][3];
  aafDstMatrix[2][0] = aafTmp[2][0]; aafDstMatrix[2][1] = aafTmp[2][1]; aafDstMatrix[2][2] = aafTmp[2][2]; aafDstMatrix[2][3] = aafTmp[2][3];
  aafDstMatrix[3][0] = aafTmp[3][0]; aafDstMatrix[3][1] = aafTmp[3][1]; aafDstMatrix[3][2] = aafTmp[3][2]; aafDstMatrix[3][3] = aafTmp[3][3];
}

#ifdef _ASM_SSE_
void MMMulSSE(float aafDstMatrix[4][4], float aafSrcMatrix[4][4])
{
  static float _sse aafTmp[4][4];
  uint32_t ebx, edi;
  asm volatile
  (
#if 0
    // This is the original version :
    // Load the source matrix into register
    //                                   XPOS - F0 F1 F2 F3
    "movaps   (%%esi), %%xmm4     \n" // xmm4 = src[00 01 02 03]
    "movaps 16(%%esi), %%xmm5     \n" // xmm5 = src[10 11 12 13]
    "movaps 32(%%esi), %%xmm6     \n" // xmm6 = src[20 21 22 23]
    "movaps 48(%%esi), %%xmm7     \n" // xmm7 = src[30 31 32 33]
    // Multiply the row0
    "movss    (%%edi), %%xmm0     \n" // xmm0 = dst[00 XX XX XX]
    "movss   4(%%edi), %%xmm1     \n" // xmm1 = dst[01 XX XX XX]
    "movss   8(%%edi), %%xmm2     \n" // xmm2 = dst[02 XX XX XX]
    "movss  12(%%edi), %%xmm3     \n" // xmm3 = dst[03 XX XX XX]
    "shufps $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[00 00 00 00]
    "shufps $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[01 01 01 01]
    "shufps $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[02 02 02 02]
    "shufps $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[03 03 03 03]
    "mulps  %%xmm4, %%xmm0        \n" // xmm0 = dst[00 00 00 00] X src[00 01 02 03]
    "mulps  %%xmm5, %%xmm1        \n" // xmm1 = dst[01 01 01 01] X src[10 11 12 13]
    "mulps  %%xmm6, %%xmm2        \n" // xmm2 = dst[02 02 02 02] X src[20 21 22 23]
    "mulps  %%xmm7, %%xmm3        \n" // xmm3 = dst[03 03 03 03] X src[30 31 32 33]
    "addps  %%xmm1, %%xmm0        \n" // xmm0 = xmm0 + xmm1
    "addps  %%xmm3, %%xmm2        \n" // xmm2 = xmm2 + xmm3
    "addps  %%xmm2, %%xmm0        \n" // xmm0 = dst[row0] X src
    "movaps %%xmm0, (%%ebx)       \n" // save row0 result to temporary matrix
    // Multiply the row1
    "movss  16(%%edi), %%xmm0     \n" // xmm0 = dst[10 XX XX XX]
    "movss  20(%%edi), %%xmm1     \n" // xmm1 = dst[11 XX XX XX]
    "movss  24(%%edi), %%xmm2     \n" // xmm2 = dst[12 XX XX XX]
    "movss  28(%%edi), %%xmm3     \n" // xmm3 = dst[13 XX XX XX]
    "shufps $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[10 10 10 10]
    "shufps $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[11 11 11 11]
    "shufps $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[12 12 12 12]
    "shufps $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[13 13 13 13]
    "mulps  %%xmm4, %%xmm0        \n" // xmm0 = dst[10 10 10 10] X src[00 01 02 03]
    "mulps  %%xmm5, %%xmm1        \n" // xmm1 = dst[11 11 11 11] X src[10 11 12 13]
    "mulps  %%xmm6, %%xmm2        \n" // xmm2 = dst[12 12 12 12] X src[20 21 22 23]
    "mulps  %%xmm7, %%xmm3        \n" // xmm3 = dst[13 13 13 13] X src[30 31 32 33]
    "addps  %%xmm1, %%xmm0        \n" // xmm0 = xmm0 + xmm1
    "addps  %%xmm3, %%xmm2        \n" // xmm2 = xmm2 + xmm3
    "addps  %%xmm2, %%xmm0        \n" // xmm0 = dst[row1] X src
    "movaps %%xmm0, 16(%%ebx)     \n" // save row1 result to temporary matrix
    // Multiply the row2
    "movss  32(%%edi), %%xmm0     \n" // xmm0 = dst[20 XX XX XX]
    "movss  36(%%edi), %%xmm1     \n" // xmm1 = dst[21 XX XX XX]
    "movss  40(%%edi), %%xmm2     \n" // xmm2 = dst[22 XX XX XX]
    "movss  44(%%edi), %%xmm3     \n" // xmm3 = dst[23 XX XX XX]
    "shufps $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[20 20 20 20]
    "shufps $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[21 21 21 21]
    "shufps $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[22 22 22 22]
    "shufps $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[23 23 23 23]
    "mulps  %%xmm4, %%xmm0        \n" // xmm0 = dst[20 20 20 20] X src[00 01 02 03]
    "mulps  %%xmm5, %%xmm1        \n" // xmm1 = dst[21 21 21 21] X src[10 11 12 13]
    "mulps  %%xmm6, %%xmm2        \n" // xmm2 = dst[22 22 22 22] X src[20 21 22 23]
    "mulps  %%xmm7, %%xmm3        \n" // xmm3 = dst[23 23 23 23] X src[30 31 32 33]
    "addps  %%xmm1, %%xmm0        \n" // xmm0 = xmm0 + xmm1
    "addps  %%xmm3, %%xmm2        \n" // xmm2 = xmm2 + xmm3
    "addps  %%xmm2, %%xmm0        \n" // xmm0 = dst[row2] X src
    "movaps %%xmm0, 32(%%ebx)     \n" // save row2 result to temporary matrix
    // Multiply the row3
    "movss  48(%%edi), %%xmm0     \n" // xmm0 = dst[30 XX XX XX]
    "movss  52(%%edi), %%xmm1     \n" // xmm1 = dst[31 XX XX XX]
    "movss  56(%%edi), %%xmm2     \n" // xmm2 = dst[32 XX XX XX]
    "movss  60(%%edi), %%xmm3     \n" // xmm3 = dst[33 XX XX XX]
    "shufps $0x00, %%xmm0, %%xmm0 \n" // xmm0 = dst[30 30 30 30]
    "shufps $0x00, %%xmm1, %%xmm1 \n" // xmm1 = dst[31 31 31 31]
    "shufps $0x00, %%xmm2, %%xmm2 \n" // xmm2 = dst[32 32 32 32]
    "shufps $0x00, %%xmm3, %%xmm3 \n" // xmm3 = dst[33 33 33 33]
    "mulps  %%xmm4, %%xmm0        \n" // xmm0 = dst[30 30 30 30] X src[00 01 02 03]
    "mulps  %%xmm5, %%xmm1        \n" // xmm1 = dst[31 31 31 31] X src[10 11 12 13]
    "mulps  %%xmm6, %%xmm2        \n" // xmm2 = dst[32 32 32 32] X src[20 21 22 23]
    "mulps  %%xmm7, %%xmm3        \n" // xmm3 = dst[33 33 33 33] X src[30 31 32 33]
    "addps  %%xmm1, %%xmm0        \n" // xmm0 = xmm0 + xmm1
    "addps  %%xmm3, %%xmm2        \n" // xmm2 = xmm2 + xmm3
    "addps  %%xmm2, %%xmm0        \n" // xmm0 = dst[row3] X src
    "movaps %%xmm0, 48(%%ebx)     \n" // save row3 result to temporary matrix
    // Copy temporary matrix to destination matrix
    "movaps   (%%ebx), %%xmm0     \n" // xmm0 = tmp0
    "movaps 16(%%ebx), %%xmm1     \n" // xmm1 = tmp1
    "movaps 32(%%ebx), %%xmm2     \n" // xmm2 = tmp2
    "movaps 48(%%ebx), %%xmm3     \n" // xmm3 = tmp3
    "movaps %%xmm0,   (%%edi)     \n" // dst0 = tmp0
    "movaps %%xmm1, 16(%%edi)     \n" // dst1 = tmp1
    "movaps %%xmm2, 32(%%edi)     \n" // dst2 = tmp2
    "movaps %%xmm3, 48(%%edi)     \n" // dst3 = tmp3
#else
    // This is the optimized version (runs slightly faster than the original version) :
    // Load the source matrix into register
    //                                      XPOS - F0 F1 F2 F3
    "movaps   (%%esi), %%xmm4         \n"// xmm4 = src[00 01 02 03]
    "movaps 16(%%esi), %%xmm5         \n"// xmm5 = src[10 11 12 13]
    "movaps 32(%%esi), %%xmm6         \n"// xmm6 = src[20 21 22 23]
    "movaps 48(%%esi), %%xmm7         \n"// xmm7 = src[30 31 32 33]
    // Multiply the row0
    "movss    (%%edi), %%xmm0         \n"// xmm0 = dst[00 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n"// xmm0 = dst[00 00 00 00]
    "movss   4(%%edi), %%xmm1         \n"// xmm1 = dst[01 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n"// xmm1 = dst[01 01 01 01]
    "movss   8(%%edi), %%xmm2         \n"// xmm2 = dst[02 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n"// xmm2 = dst[02 02 02 02]
    "movss  12(%%edi), %%xmm3         \n"// xmm3 = dst[03 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n"// xmm3 = dst[03 03 03 03]
    "mulps     %%xmm4, %%xmm0         \n"// xmm0 = dst[00 00 00 00] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n"// xmm1 = dst[01 01 01 01] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n"// xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n"// xmm2 = dst[02 02 02 02] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n"// xmm3 = dst[03 03 03 03] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n"// xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n"// xmm0 = dst[row0] X src
    "movaps    %%xmm0, (%%ebx)        \n"// save row0 result to temporary matrix
    // Multiply the row1
    "movss  16(%%edi), %%xmm0         \n"// xmm0 = dst[10 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n"// xmm0 = dst[10 10 10 10]
    "movss  20(%%edi), %%xmm1         \n"// xmm1 = dst[11 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n"// xmm1 = dst[11 11 11 11]
    "movss  24(%%edi), %%xmm2         \n"// xmm2 = dst[12 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n"// xmm2 = dst[12 12 12 12]
    "movss  28(%%edi), %%xmm3         \n"// xmm3 = dst[13 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n"// xmm3 = dst[13 13 13 13]
    "mulps     %%xmm4, %%xmm0         \n"// xmm0 = dst[10 10 10 10] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n"// xmm1 = dst[11 11 11 11] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n"// xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n"// xmm2 = dst[12 12 12 12] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n"// xmm3 = dst[13 13 13 13] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n"// xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n"// xmm0 = dst[row1] X src
    "movaps    %%xmm0, 16(%%ebx)      \n"// save row1 result to temporary matrix
    // Multiply the row2
    "movss  32(%%edi), %%xmm0         \n"// xmm0 = dst[20 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n"// xmm0 = dst[20 20 20 20]
    "movss  36(%%edi), %%xmm1         \n"// xmm1 = dst[21 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n"// xmm1 = dst[21 21 21 21]
    "movss  40(%%edi), %%xmm2         \n"// xmm2 = dst[22 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n"// xmm2 = dst[22 22 22 22]
    "movss  44(%%edi), %%xmm3         \n"// xmm3 = dst[23 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n"// xmm3 = dst[23 23 23 23]
    "mulps     %%xmm4, %%xmm0         \n"// xmm0 = dst[20 20 20 20] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n"// xmm1 = dst[21 21 21 21] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n"// xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n"// xmm2 = dst[22 22 22 22] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n"// xmm3 = dst[23 23 23 23] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n"// xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n"// xmm0 = dst[row2] X src
    "movaps    %%xmm0, 32(%%ebx)      \n"// save row2 result to temporary matrix
    // Multiply the row3
    "movss  48(%%edi), %%xmm0         \n"// xmm0 = dst[30 XX XX XX]
    "shufps     $0x00, %%xmm0, %%xmm0 \n"// xmm0 = dst[30 30 30 30]
    "movss  52(%%edi), %%xmm1         \n"// xmm1 = dst[31 XX XX XX]
    "shufps     $0x00, %%xmm1, %%xmm1 \n"// xmm1 = dst[31 31 31 31]
    "movss  56(%%edi), %%xmm2         \n"// xmm2 = dst[32 XX XX XX]
    "shufps     $0x00, %%xmm2, %%xmm2 \n"// xmm2 = dst[32 32 32 32]
    "movss  60(%%edi), %%xmm3         \n"// xmm3 = dst[33 XX XX XX]
    "shufps     $0x00, %%xmm3, %%xmm3 \n"// xmm3 = dst[33 33 33 33]
    "mulps     %%xmm4, %%xmm0         \n"// xmm0 = dst[30 30 30 30] X src[00 01 02 03]
    "mulps     %%xmm5, %%xmm1         \n"// xmm1 = dst[31 31 31 31] X src[10 11 12 13]
    "addps     %%xmm1, %%xmm0         \n"// xmm0 = xmm0 + xmm1
    "mulps     %%xmm6, %%xmm2         \n"// xmm2 = dst[32 32 32 32] X src[20 21 22 23]
    "mulps     %%xmm7, %%xmm3         \n"// xmm3 = dst[33 33 33 33] X src[30 31 32 33]
    "addps     %%xmm3, %%xmm2         \n"// xmm2 = xmm2 + xmm3
    "addps     %%xmm2, %%xmm0         \n"// xmm0 = dst[row3] X src
    // Copy temporary matrix to destination matrix
    "movaps   (%%ebx), %%xmm1         \n"// xmm0 = tmp0
    "movaps 16(%%ebx), %%xmm2         \n"// xmm1 = tmp1
    "movaps 32(%%ebx), %%xmm3         \n"// xmm2 = tmp2
    "movaps %%xmm1,   (%%edi)         \n"// dst0 = tmp0
    "movaps %%xmm2, 16(%%edi)         \n"// dst1 = tmp1
    "movaps %%xmm3, 32(%%edi)         \n"// dst2 = tmp2
    "movaps %%xmm0, 48(%%edi)         \n"// dst3 = tmp3
  #endif
    : "=&b" (ebx),  "=&D" (edi)                              \
    :   "0" (aafTmp), "1" (aafDstMatrix), "S" (aafSrcMatrix) \
    : "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7", "memory"
  );
}
#endif


//
// Helper functions
//

// Timer
inline uint64_t GetCycles()
{
  uint64_t v;
  asm volatile ("rdtsc" : "=A" (v));
  return(v);
}

// Logger
void Log(const char *pcMsg, ...)
{
  // Temporary variables
  char    acBuff[1024];
  va_list vlTmp;

  // Extract the pcMsg
  va_start(vlTmp, pcMsg);
    vsnprintf(acBuff, sizeof(acBuff), pcMsg, vlTmp);
  va_end(vlTmp);

  // Print the message
  printf(acBuff);
}



//
// Main program (benchmaker)
//
#define TOT_ITER   (1024 * 1024 * 64)
#define TOT_ITER_4 (TOT_ITER / 4)
#define DIV_FACT_4 (TOT_ITER / TOT_ITER_4)

int main()
{
  // Aligned local variables
  float _sse afMatS[4][4] = { { 00,01,02,03 }, { 10,11,12,13 }, { 20,21,22,23 }, { 30,31,32,33 } };
  float _sse afVecS[4]    = { 10, 20, 30, 1 };
  float _sse afMat1[4][4];
  float _sse afMat2[4][4];
  float _sse afVec1[4]    = { 0, 0, 0, 0 };

  // For benchmarking
  uint64_t un64TS, un64TE, un64TC, un64TD;
  uint32_t un32Cnt;
  double   afSpd[2];

  // Display cycles / seconds
  un64TS = GetCycles(); sleep(0); un64TE = GetCycles();
  un64TD = un64TE - un64TS;
  un64TS = GetCycles(); sleep(1); un64TE = GetCycles();
  Log("\n%llu cycles per second\n", un64TE - un64TS - un64TD);

  // Calculate overhead
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++);
  un64TE = GetCycles();
  un64TD = un64TE - un64TS;
  Log("\nLooped timing overhead = %llu cycles\n", un64TD);

  // Benchmark matrix copy functions
  Log("\nBenchmarking MatrixCopy functions ...\n");
  Log( "  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMatS[0][0], afMatS[0][1], afMatS[0][2], afMatS[0][3], afMatS[1][0], afMatS[1][1], afMatS[1][2], afMatS[1][3],
    afMatS[2][0], afMatS[2][1], afMatS[2][2], afMatS[2][3], afMatS[3][0], afMatS[3][1], afMatS[3][2], afMatS[3][3]);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) MCopy(afMat1, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);
  afSpd[0] = double(un64TC) / double(TOT_ITER);
  Log("  C++ copy : %7.5f cycles\n", afSpd[0]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#ifdef _ASM_MMX_
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) MCopyMMX(afMat1, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);
  afSpd[1] = double(un64TC) / double(TOT_ITER);
  Log("  MMX copy : %7.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#else
  Log("  MMX copy : <<< MMX is not enabled >>>\n");
#endif
#ifdef _ASM_SSE_
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) MCopySSE(afMat1, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);
  afSpd[1] = double(un64TC) / double(TOT_ITER);
  Log("  SSE copy : %7.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#else
  Log("  SSE copy : <<< SSE is not enabled >>>\n");
#endif

  // Benchmark matrix identity functions
  Log("\nBenchmarking MatrixIdentity functions ...\n");
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) M1(afMat1);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[0] = double(un64TC) / double(TOT_ITER);
  Log("  C++ identity : %7.5f cycles\n", afSpd[0]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#ifdef _ASM_MMX_
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) M1MMX(afMat1);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[1] = double(un64TC) / double(TOT_ITER);
  Log("  MMX identity : %7.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#else
  Log("  MMX identity : <<< MMX is not enabled >>>\n");
#endif
#ifdef _ASM_SSE_
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) M1SSE(afMat1);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[1] = double(un64TC) / double(TOT_ITER);
  Log("  SSE identity : %7.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#else
  Log("  SSE identity : <<< SSE is not enabled >>>\n");
#endif

  // Benchmark matrix transpose functions
  Log("\nBenchmarking MatrixTranspose functions ...\n");
  Log( "  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMatS[0][0], afMatS[0][1], afMatS[0][2], afMatS[0][3], afMatS[1][0], afMatS[1][1], afMatS[1][2], afMatS[1][3],
    afMatS[2][0], afMatS[2][1], afMatS[2][2], afMatS[2][3], afMatS[3][0], afMatS[3][1], afMatS[3][2], afMatS[3][3]);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) MTranspose(afMat1, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[0] = double(un64TC) / double(TOT_ITER);
  Log("  C++ transpose : %7.5f cycles\n", afSpd[0]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#ifdef _ASM_SSE_
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) MTransposeSSE(afMat1, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[1] = double(un64TC) / double(TOT_ITER);
  Log("  SSE transpose : %7.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#else
  Log("  SSE transpose : <<< SSE is not enabled >>>\n");
#endif

  // Benchmark vector-matrix multiplication functions
  Log("\nBenchmarking VectorMatrixMultiplication functions ...\n");
  Log( "  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n", afVecS[0], afVecS[1], afVecS[2], afVecS[3]);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) VMMul(afVec1, afVecS, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[0] = double(un64TC) / double(TOT_ITER);
  Log("  C++ multiplication : %8.5f cycles\n", afSpd[0]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n", afVec1[0], afVec1[1], afVec1[2], afVec1[3]);
#ifdef _ASM_SSE_
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) VMMulSSE(afVec1, afVecS, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[1] = double(un64TC) / double(TOT_ITER);
  Log("  SSE multiplication : %8.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n", afVec1[0], afVec1[1], afVec1[2], afVec1[3]);
#else
  Log("  SSE multiplication : <<< SSE is not enabled >>>\n");
#endif
#ifdef _ASM_3DNOW_
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER; un32Cnt++) VMMul3DNow(afVec1, afVecS, afMatS);
  un64TE = GetCycles();
  un64TC = ((un64TE - un64TS) > un64TD) ? (un64TE - un64TS - un64TD) : (un64TE - un64TS);;
  afSpd[1] = double(un64TC) / double(TOT_ITER);
  Log("  3DN multiplication : %8.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  Log( "    |   %6.2f   %6.2f   %6.2f   %6.2f   |\n", afVec1[0], afVec1[1], afVec1[2], afVec1[3]);
#else
  Log("  3DN multiplication : <<< 3DNow! is not enabled >>>\n");
#endif

  // Benchmark matrix-matrix multiplication functions
  Log("\nBenchmarking MatrixMatrixMultiplication functions ...\n");
  Log( "  |   %6.2f   %6.2f   %6.2f   %6.2f   | 2\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n  |   %6.2f   %6.2f   %6.2f   %6.2f   |\n",
    afMatS[0][0], afMatS[0][1], afMatS[0][2], afMatS[0][3], afMatS[1][0], afMatS[1][1], afMatS[1][2], afMatS[1][3],
    afMatS[2][0], afMatS[2][1], afMatS[2][2], afMatS[2][3], afMatS[3][0], afMatS[3][1], afMatS[3][2], afMatS[3][3]);
  M1(afMat2);
  MCopy(afMat1, afMatS);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER_4; un32Cnt++) MMMul(afMat1, afMat2);
  un64TE = GetCycles();
  un64TC = un64TE - un64TS - (un64TD / DIV_FACT_4);
  afSpd[0] = double(un64TC) / double(TOT_ITER_4);
  Log("  C++ multiplication : %8.5f cycles\n", afSpd[0]);
  MMMul(afMat1, afMatS);
  Log( "    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#ifdef _ASM_SSE_
  MCopySSE(afMat1, afMatS);
  un64TS = GetCycles();
  for(un32Cnt = 0; un32Cnt < TOT_ITER_4; un32Cnt++) MMMulSSE(afMat1, afMat2);
  un64TE = GetCycles();
  un64TC = un64TE - un64TS - (un64TD / DIV_FACT_4);
  afSpd[1] = double(un64TC) / double(TOT_ITER_4);
  Log("  SSE multiplication : %8.5f cycles (%4.2fX faster)\n", afSpd[1], afSpd[0] / afSpd[1]);
  MMMulSSE(afMat1, afMatS);
  Log( "    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n    |   %7.2f   %7.2f   %7.2f   %7.2f   |\n",
    afMat1[0][0], afMat1[0][1], afMat1[0][2], afMat1[0][3], afMat1[1][0], afMat1[1][1], afMat1[1][2], afMat1[1][3],
    afMat1[2][0], afMat1[2][1], afMat1[2][2], afMat1[2][3], afMat1[3][0], afMat1[3][1], afMat1[3][2], afMat1[3][3]);
#else
  Log("  SSE multiplication : <<< SSE is not enabled >>>\n");
#endif
  Log("\n");

  // End
  return(0);
}
