// ############################################################################################
//  mcatedit-MCatEdit-StdEH.cc
//  Implementation of mcatedit's CMCatEdit and its dialog page classes
//  (standard event handlers)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// mcatedit's headers
#include "mcatedit-MCatEdit.h"


//
// Our dialog page: Directory Selection
//
void CPage_SelDir::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);
       if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER) _onBtnOK(NULL);
  else if(keySymbol==VKS_ESC)                              _onBtnCancel(NULL);
}




//
// Our dialog page: Locale Management
//
void CPage_MLoc::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);
  if(keySymbol==VKS_ENTER    ||
     keySymbol==VKS_KP_ENTER ||
     keySymbol==VKS_ESC) _onBtnDone(NULL);
}




//
// Our dialog page: Find in ID or MT
//
void CPage_Find::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);
  if(keySymbol==VKS_ENTER || keySymbol==VKS_KP_ENTER) _onBtnSearch(NULL);
  if(keySymbol==VKS_ESC) _onBtnDone(NULL);
}




//
// Our dialog page: Main Page
//
void CPage_Main::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);
  if(keySymbol == VKS_ESC) _app.EventCloseWindow();
}

void CPage_Main::DisplayTooltip(const char *tooltip)
{
  stbSB.SetString(tooltip);
  stbSB.DrawControl();
}




//
// Our application class : standard event handlers
//
void CMCatEdit::EventUpdateFrame(bool bWindowIsActive)
{
  CMCatEditBase::EventUpdateFrame(bWindowIsActive);

  if(bWindowIsActive)
    usleep(1000);
  else {
    csVStrg.RestoreCurrentView(APP_INACTIVE_STORE);
    iVideoEnv().UpdateDisplay();
  }
}

void CMCatEdit::EventWindowActivated()
{
  CMCatEditBase::EventWindowActivated();

  iVideoEnv().UpdateDisplay();
}

void CMCatEdit::EventWindowDeactivated()
{
  CMCatEditBase::EventWindowActivated();

  iVideoEnv().SetForegroundColor(iVideoEnv().CreateColor(255, 255, 255));
  iVideoEnv().UseDefaultBigFont();
  iVideoEnv().DrawText(0, 150, "Inactive State", ETA_CenterScreen);
  iVideoEnv().UseDefaultSmallFont();
  iVideoEnv().UpdateDisplay();

  csVStrg.StoreCurrentView(APP_INACTIVE_STORE);
}

void CMCatEdit::EventCloseWindow()
{
  csVStrg.StoreCurrentView(PRE_DIALOG_STORE);

  iVideoEnv().ApplyBlackScreen();
  if(iVideoEnv().MessageBox(_M("Exit Editor"),
                            _M("Do you really want to exit ?"),
                            EMBI_Question,
                            EMBB_YesNo,
                            EMBS_Small) == EMBRV_Yes)
    CMCatEditBase::EventCloseWindow();

  csVStrg.RestoreCurrentView(PRE_DIALOG_STORE);
}
