// ############################################################################################
//  roomedit-DlgPages-Main-StdEH.cc
//  Implementation of roomedit's CPage_Main class (standard event handlers)
//  -------------------------------------------------------------------------------------------
//  Copyright (C) 2004-2006 Aloysius Indrayanto
//  This file is part of VortexGE
//
//  This program is free software; you can redistribute it and/or modify it under the terms of
//  the GNU Lesser General Public License as published by the Free Software Foundation; either
//  version 2.1 of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU Lesser General Public License for more details.
//
//  You should have received a copy of the GNU Lesser General Public License along with this
//  library; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
//  Boston, MA 02111-1307 USA.
// ############################################################################################

// roomedit's headers
#include "roomedit-DlgPages-Main.h"


//
// Our main dialog page: standard event handlers
//
void CPage_Main::HandleEvent_OnKeyPress(EKeySymbol keySymbol, const char *str)
{
  // Handle keyboard shortcuts
  if(HandleKeyboardShortcut(keySymbol, str, _keyMap.ctrl, _keyMap.alt, _keyMap.shift)) return;

  // Handle movements
  _keyMap.ctrl  |= (keySymbol == VKS_LCTRL)  || (keySymbol == VKS_RCTRL);
  _keyMap.alt   |= (keySymbol == VKS_LALT)   || (keySymbol == VKS_RALT);
  _keyMap.shift |= (keySymbol == VKS_LSHIFT) || (keySymbol == VKS_RSHIFT);
  _keyMap.up    |= (keySymbol == VKS_UP)     || (keySymbol == VKS_KP_8 );
  _keyMap.down  |= (keySymbol == VKS_DOWN)   || (keySymbol == VKS_KP_2 );
  _keyMap.left  |= (keySymbol == VKS_LEFT)   || (keySymbol == VKS_KP_4 );
  _keyMap.right |= (keySymbol == VKS_RIGHT)  || (keySymbol == VKS_KP_6 );
  _keyMap.home  |= (keySymbol == VKS_HOME)   || (keySymbol == VKS_KP_7 );
  _keyMap.pgup  |= (keySymbol == VKS_PGUP)   || (keySymbol == VKS_KP_9 );
}

void CPage_Main::HandleEvent_OnKeyRelease(EKeySymbol keySymbol, const char *str)
{
  UNUSED_PARAM(str);

  _keyMap.ctrl  &= !((keySymbol == VKS_LCTRL)  || (keySymbol == VKS_RCTRL));
  _keyMap.alt   &= !((keySymbol == VKS_LALT)   || (keySymbol == VKS_RALT));
  _keyMap.shift &= !((keySymbol == VKS_LSHIFT) || (keySymbol == VKS_RSHIFT));
  _keyMap.up    &= !((keySymbol == VKS_UP)     || (keySymbol == VKS_KP_8 ));
  _keyMap.down  &= !((keySymbol == VKS_DOWN)   || (keySymbol == VKS_KP_2 ));
  _keyMap.left  &= !((keySymbol == VKS_LEFT)   || (keySymbol == VKS_KP_4 ));
  _keyMap.right &= !((keySymbol == VKS_RIGHT)  || (keySymbol == VKS_KP_6 ));
  _keyMap.home  &= !((keySymbol == VKS_HOME)   || (keySymbol == VKS_KP_7 ));
  _keyMap.pgup  &= !((keySymbol == VKS_PGUP)   || (keySymbol == VKS_KP_9 ));
}

void CPage_Main::HandleEvent_OnUpdateFrame(bool windowIsActive)
{
  if(windowIsActive) SmartUpdateAll();
}

void CPage_Main::DisplayTooltip(const char *tooltip)
{
  stbSB.SetString(tooltip);
  stbSB.DrawControl();
}

void CPage_Main::CustomDraw_OnRedrawAllControls()
{
  _iVideoEnv->PutRenderImageOnDisplay(rbpX, rbpY);
}

void CPage_Main::CustomDraw_OnButtonPress(EBtnSymbol eButtonSym)
{
  UNUSED_PARAM(eButtonSym);
  _iVideoEnv->PutRenderImageOnDisplay(rbpX, rbpY);
}

void CPage_Main::CustomDraw_OnButtonRelease(EBtnSymbol eButtonSym)
{
  UNUSED_PARAM(eButtonSym);
  _iVideoEnv->PutRenderImageOnDisplay(rbpX, rbpY);
}

void CPage_Main::CustomDraw_OnMouseMove(int32_t pointerXPos, int32_t pointerYPos)
{ 
  UNUSED_PARAMS_2(pointerXPos, pointerYPos); 
  _iVideoEnv->PutRenderImageOnDisplay(rbpX, rbpY);
}
